/*
main.c
Copyright © 2008 William Astle

This file is part of LWASM.

LWASM is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.


Implements the program startup code

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <argp.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "lwasm.h"

// external declarations
extern void lwasm_pass1(asmstate_t *as);
extern void lwasm_pass2(asmstate_t *as);
extern void lwasm_list(asmstate_t *as);
extern void lwasm_output(asmstate_t *as);
extern void pseudo_pragma_real(asmstate_t *as, lwasm_line_t *cl, char **optr, int error);

// command line option handling
const char *argp_program_version = "LWASM from " PACKAGE_STRING;
const char *argp_program_bug_address = PACKAGE_BUGREPORT;

static error_t parse_opts(int key, char *arg, struct argp_state *state)
{
	asmstate_t *as = state -> input;
	char *p;
	
	switch (key)
	{
	case 'o':
		// output
		if (as -> outfile)
		{
		}
		as -> outfile = arg;
		break;
	
	case 'd':
		// debug
		debug_level++;
		break;
	
	case 'l':
		// list
		if (arg)
			as -> listfile = arg;
		else
			as -> listfile = "-";
		break;
	
	case 'b':
		// decb output
		as -> outformat = OUTPUT_DECB;
		break;
	
	case 'r':
		// raw binary output
		as -> outformat = OUTPUT_RAW;
		break;
	
	case 0x100:
		// proprietary object format
		as -> outformat = OUTPUT_OBJ;
		break;
		
	case 'f':
		// output format
		if (!strcasecmp(arg, "decb"))
			as -> outformat = OUTPUT_DECB;
		else if (!strcasecmp(arg, "raw"))
			as -> outformat = OUTPUT_RAW;
		else if (!strcasecmp(arg, "obj"))
			as -> outformat = OUTPUT_OBJ;
		else
		{
			fprintf(stderr, "Invalid output format: %s\n", arg);
			exit(1);
		}
		break;
	
	case 'p':
		// pragmas
		p = arg;
		pseudo_pragma_real(as, NULL, &p, 2);
		if (!p)
		{
			fprintf(stderr, "Invalid pragma string: %s\n", arg);
			exit(1);
		}
		break;

	case ARGP_KEY_END:
		// done; sanity check
		if (!as -> outfile)
			as -> outfile = "a.out";
		break;
	
	case ARGP_KEY_ARG:
		// non-option arg
		if (as -> infile)
			argp_usage(state);
		as -> infile = arg;
		break;
		
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static struct argp_option options[] =
{
	{ "output",		'o',	"FILE",	0,
				"Output to FILE"},
	{ "debug",		'd',	0,		0,
				"Set debug mode"},
	{ "format",		'f',	"TYPE",	0,
				"Select output format: decb, raw, obj"},
	{ "list",		'l',	"FILE",	OPTION_ARG_OPTIONAL,
				"Generate list [to FILE]"},
	{ "decb",		'b',	0,		0,
				"Generate DECB .bin format output, equivalent of --format=decb"},
	{ "raw",		'r',	0,		0,
				"Generate raw binary format output, equivalent of --format=raw"},
	{ "obj",		0x100,		0,		0,
				"Generate proprietary object file format for later linking, equivalent of --format=obj" },
	{ "pragma",		'p',	"PRAGMA",	0,
				"Set an assembler pragma to any value understood by the \"pragma\" pseudo op"},
	{ 0 }
};

static struct argp argp =
{
	options,
	parse_opts,
	"<input file>",
	"LWASM, a HD6309 and MC6809 cross-assembler"
};

// main function; parse command line, set up assembler state, and run the
// assembler on the first file
int main(int argc, char **argv)
{
	// assembler state
	asmstate_t asmstate = { 0 };

	argp_parse(&argp, argc, argv, 0, 0, &asmstate);

	if (!asmstate.infile)
	{
		fprintf(stderr, "No input files specified.\n");
		exit(1);
	}	

	/* pass 1 - collect the symbols and assign addresses where possible */
	/* pass 1 also resolves included files, etc. */
	/* that means files are read exactly once unless included multiple times */
	lwasm_pass1(&asmstate);
	
	// pass 2: actually generate the code; if any phasing errors appear
	// at this stage, we have a bug
	lwasm_pass2(&asmstate);

	// now make a pretty listing
	lwasm_list(&asmstate);

	// now write the code out to the output file
	lwasm_output(&asmstate);

	if (asmstate.errorcount > 0)
		exit(1);

	exit(0);
}

